import unittest
from unittest.mock import MagicMock, patch

from flask import Flask
from mlflow.exceptions import MlflowException
from mlflow.protos.databricks_pb2 import BAD_REQUEST, INVALID_PARAMETER_VALUE, RESOURCE_DOES_NOT_EXIST

from mlflow_oidc_auth.permissions import Permission
from mlflow_oidc_auth.utils import (
    PermissionResult,
    can_manage_experiment,
    can_manage_registered_model,
    can_read_experiment,
    can_read_registered_model,
    check_experiment_permission,
    check_prompt_permission,
    check_registered_model_permission,
    check_admin_permission,
    effective_experiment_permission,
    effective_prompt_permission,
    effective_registered_model_permission,
    fetch_all_experiments,
    fetch_all_prompts,
    fetch_all_registered_models,
    fetch_experiments_paginated,
    fetch_registered_models_paginated,
    fetch_readable_experiments,
    fetch_readable_registered_models,
    get_experiment_id,
    get_is_admin,
    get_model_name,
    get_optional_request_param,
    get_permission_from_store_or_default,
    get_request_param,
    get_url_param,
    get_optional_url_param,
    get_username,
    _get_registered_model_permission_from_regex,
    _get_experiment_permission_from_regex,
    _get_registered_model_group_permission_from_regex,
    _get_experiment_group_permission_from_regex,
    _permission_prompt_sources_config,
    _permission_experiment_sources_config,
    _permission_registered_model_sources_config,
    _experiment_id_from_name,
)


class TestUtils(unittest.TestCase):
    def setUp(self):
        self.app = Flask(__name__)
        self.app.config["TESTING"] = True
        self.app_context = self.app.app_context()
        self.app_context.push()
        self.client = self.app.test_client()

    def tearDown(self):
        self.app_context.pop()

    @patch("mlflow_oidc_auth.utils.store")
    @patch("mlflow_oidc_auth.utils.get_username")
    def test_get_is_admin(self, mock_get_username, mock_store):
        with self.app.test_request_context():
            mock_get_username.return_value = "user"
            mock_store.get_user.return_value.is_admin = True
            self.assertTrue(get_is_admin())
            mock_store.get_user.return_value.is_admin = False
            self.assertFalse(get_is_admin())

    @patch("mlflow_oidc_auth.utils.store")
    @patch("mlflow_oidc_auth.utils.config")
    @patch("mlflow_oidc_auth.utils.get_permission")
    def test_get_permission_from_store_or_default(self, mock_get_permission, mock_config, mock_store):
        with self.app.test_request_context():
            mock_store_permission_user_func = MagicMock()
            mock_store_permission_group_func = MagicMock()
            mock_store_permission_user_func.return_value = "user_perm"
            mock_store_permission_group_func.return_value = "group_perm"
            mock_get_permission.return_value = Permission(name="perm", priority=1, can_read=True, can_update=True, can_delete=True, can_manage=True)
            mock_config.PERMISSION_SOURCE_ORDER = ["user", "group"]
            mock_config.DEFAULT_MLFLOW_PERMISSION = "default_perm"

            # user permission found
            result = get_permission_from_store_or_default({"user": mock_store_permission_user_func, "group": mock_store_permission_group_func})
            self.assertTrue(result.permission.can_manage)
            self.assertEqual(result.type, "user")

            # user not found, group found
            mock_store_permission_user_func.side_effect = MlflowException("", RESOURCE_DOES_NOT_EXIST)
            result = get_permission_from_store_or_default({"user": mock_store_permission_user_func, "group": mock_store_permission_group_func})
            self.assertTrue(result.permission.can_manage)
            self.assertEqual(result.type, "group")

            # both not found, fallback to default
            mock_store_permission_group_func.side_effect = MlflowException("", RESOURCE_DOES_NOT_EXIST)
            result = get_permission_from_store_or_default({"user": mock_store_permission_user_func, "group": mock_store_permission_group_func})
            self.assertTrue(result.permission.can_manage)
            self.assertEqual(result.type, "fallback")

            # invalid source in config
            mock_config.PERMISSION_SOURCE_ORDER = ["invalid"]
            # Just call and check fallback, don't assert logs
            result = get_permission_from_store_or_default({"user": mock_store_permission_user_func, "group": mock_store_permission_group_func})
            self.assertEqual(result.type, "fallback")

    @patch("mlflow_oidc_auth.utils.store")
    @patch("mlflow_oidc_auth.utils.get_permission_from_store_or_default")
    def test_can_manage_experiment(self, mock_get_permission_from_store_or_default, mock_store):
        with self.app.test_request_context():
            mock_get_permission_from_store_or_default.return_value = PermissionResult(
                Permission(name="perm", priority=1, can_read=True, can_update=True, can_delete=True, can_manage=True), "user"
            )
            self.assertTrue(can_manage_experiment("exp_id", "user"))
            mock_get_permission_from_store_or_default.return_value = PermissionResult(
                Permission(name="perm", priority=1, can_read=True, can_update=True, can_delete=True, can_manage=False), "user"
            )
            self.assertFalse(can_manage_experiment("exp_id", "user"))

    @patch("mlflow_oidc_auth.utils.store")
    @patch("mlflow_oidc_auth.utils.get_permission_from_store_or_default")
    def test_can_manage_registered_model(self, mock_get_permission_from_store_or_default, mock_store):
        with self.app.test_request_context():
            mock_get_permission_from_store_or_default.return_value = PermissionResult(
                Permission(name="perm", priority=1, can_read=True, can_update=True, can_delete=True, can_manage=True), "user"
            )
            self.assertTrue(can_manage_registered_model("model_name", "user"))
            mock_get_permission_from_store_or_default.return_value = PermissionResult(
                Permission(name="perm", priority=1, can_read=True, can_update=True, can_delete=True, can_manage=False), "user"
            )
            self.assertFalse(can_manage_registered_model("model_name", "user"))

    @patch("mlflow_oidc_auth.utils.store")
    @patch("mlflow_oidc_auth.utils.get_is_admin")
    @patch("mlflow_oidc_auth.utils.get_username")
    @patch("mlflow_oidc_auth.utils.get_experiment_id")
    @patch("mlflow_oidc_auth.utils.can_manage_experiment")
    @patch("mlflow_oidc_auth.utils.make_forbidden_response")
    def test_check_experiment_permission(
        self,
        mock_make_forbidden_response,
        mock_can_manage_experiment,
        mock_get_experiment_id,
        mock_get_username,
        mock_get_is_admin,
        mock_store,
    ):
        with self.app.test_request_context():
            mock_get_is_admin.return_value = False
            mock_get_username.return_value = "user"
            mock_get_experiment_id.return_value = "exp_id"
            mock_can_manage_experiment.return_value = False
            mock_make_forbidden_response.return_value = "forbidden"

            @check_experiment_permission
            def mock_func():
                return "success"

            self.assertEqual(mock_func(), "forbidden")

            mock_can_manage_experiment.return_value = True
            self.assertEqual(mock_func(), "success")

            # Admin always allowed
            mock_get_is_admin.return_value = True
            self.assertEqual(mock_func(), "success")

    @patch("mlflow_oidc_auth.utils.store")
    @patch("mlflow_oidc_auth.utils.get_is_admin")
    @patch("mlflow_oidc_auth.utils.get_username")
    @patch("mlflow_oidc_auth.utils.get_model_name")
    @patch("mlflow_oidc_auth.utils.can_manage_registered_model")
    @patch("mlflow_oidc_auth.utils.make_forbidden_response")
    def test_check_registered_model_permission(
        self,
        mock_make_forbidden_response,
        mock_can_manage_registered_model,
        mock_get_model_name,
        mock_get_username,
        mock_get_is_admin,
        mock_store,
    ):
        with self.app.test_request_context():
            mock_get_is_admin.return_value = False
            mock_get_username.return_value = "user"
            mock_get_model_name.return_value = "model_name"
            mock_can_manage_registered_model.return_value = False
            mock_make_forbidden_response.return_value = "forbidden"

            @check_registered_model_permission
            def mock_func():
                return "success"

            self.assertEqual(mock_func(), "forbidden")

            mock_can_manage_registered_model.return_value = True
            self.assertEqual(mock_func(), "success")

            # Admin always allowed
            mock_get_is_admin.return_value = True
            self.assertEqual(mock_func(), "success")

    @patch("mlflow_oidc_auth.utils.store")
    @patch("mlflow_oidc_auth.utils.get_is_admin")
    @patch("mlflow_oidc_auth.utils.get_username")
    @patch("mlflow_oidc_auth.utils.get_model_name")
    @patch("mlflow_oidc_auth.utils.can_manage_registered_model")
    @patch("mlflow_oidc_auth.utils.make_forbidden_response")
    def test_check_prompt_permission(
        self,
        mock_make_forbidden_response,
        mock_can_manage_registered_model,
        mock_get_model_name,
        mock_get_username,
        mock_get_is_admin,
        mock_store,
    ):
        with self.app.test_request_context():
            mock_get_is_admin.return_value = False
            mock_get_username.return_value = "user"
            mock_get_model_name.return_value = "prompt_name"
            mock_can_manage_registered_model.return_value = False
            mock_make_forbidden_response.return_value = "forbidden"

            @check_prompt_permission
            def mock_func():
                return "success"

            self.assertEqual(mock_func(), "forbidden")

            mock_can_manage_registered_model.return_value = True
            self.assertEqual(mock_func(), "success")

            # Admin always allowed
            mock_get_is_admin.return_value = True
            self.assertEqual(mock_func(), "success")

    def test_get_request_param(self):
        # GET method, param present
        with self.app.test_request_context("/?foo=bar", method="GET"):
            self.assertEqual(get_request_param("foo"), "bar")
        # POST method, param present
        with self.app.test_request_context("/", method="POST", json={"foo": "baz"}):
            self.assertEqual(get_request_param("foo"), "baz")
        # param missing, run_id fallback to run_uuid
        with self.app.test_request_context("/", method="GET"):
            with patch("mlflow_oidc_auth.utils.get_request_param", return_value="uuid_val") as mock_get:
                self.assertEqual(get_request_param("run_id"), "uuid_val")
        # param missing, not run_id
        with self.app.test_request_context("/", method="GET"):
            with self.assertRaises(MlflowException) as cm:
                get_request_param("notfound")
            self.assertEqual(cm.exception.error_code, "INVALID_PARAMETER_VALUE")
        # unsupported method
        with self.app.test_request_context("/", method="PUT"):
            with self.assertRaises(MlflowException) as cm:
                get_request_param("foo")
            self.assertEqual(cm.exception.error_code, "BAD_REQUEST")

    def test_get_optional_request_param(self):
        # GET method, param present
        with self.app.test_request_context("/?foo=bar", method="GET"):
            self.assertEqual(get_optional_request_param("foo"), "bar")
        # POST method, param present
        with self.app.test_request_context("/", method="POST", json={"foo": "baz"}):
            self.assertEqual(get_optional_request_param("foo"), "baz")
        # param missing
        with self.app.test_request_context("/", method="GET"):
            self.assertIsNone(get_optional_request_param("notfound"))
        # unsupported method
        with self.app.test_request_context("/", method="PUT"):
            with self.assertRaises(MlflowException) as cm:
                get_optional_request_param("foo")
            self.assertEqual(cm.exception.error_code, "BAD_REQUEST")

    @patch("mlflow_oidc_auth.utils._get_tracking_store")
    def test_get_experiment_id(self, mock_tracking_store):
        # GET method, experiment_id present
        with self.app.test_request_context("/?experiment_id=123", method="GET"):
            self.assertEqual(get_experiment_id(), "123")
        # POST method, experiment_id present
        with self.app.test_request_context("/", method="POST", json={"experiment_id": "456"}, content_type="application/json"):
            self.assertEqual(get_experiment_id(), "456")
        # experiment_name present
        with self.app.test_request_context("/?experiment_name=exp", method="GET"):
            mock_tracking_store().get_experiment_by_name.return_value.experiment_id = "789"
            self.assertEqual(get_experiment_id(), "789")
        # missing both - mock request.json to avoid content type issues
        with self.app.test_request_context("/", method="GET"):
            with patch("mlflow_oidc_auth.utils.request") as mock_request:
                mock_request.view_args = None
                mock_request.args = {}
                mock_request.json = None
                with self.assertRaises(MlflowException) as cm:
                    get_experiment_id()
                self.assertEqual(cm.exception.error_code, "INVALID_PARAMETER_VALUE")
        # unsupported method
        with self.app.test_request_context("/", method="PUT"):
            with patch("mlflow_oidc_auth.utils.request") as mock_request:
                mock_request.view_args = None
                mock_request.args = {}
                mock_request.json = None
                with self.assertRaises(MlflowException) as cm:
                    get_experiment_id()
                self.assertEqual(cm.exception.error_code, "INVALID_PARAMETER_VALUE")

    @patch("mlflow_oidc_auth.utils.store")
    @patch("mlflow_oidc_auth.utils.validate_token")
    def test_get_username(self, mock_validate_token, mock_store):
        with self.app.test_request_context():
            # session username
            with patch("mlflow_oidc_auth.utils.session", {"username": "session_user"}):
                with patch("mlflow_oidc_auth.utils.request") as mock_request:
                    mock_request.authorization = None
                    self.assertEqual(get_username(), "session_user")
            # basic auth username
            with patch("mlflow_oidc_auth.utils.session", {}):

                class AuthBasic:
                    type = "basic"
                    username = "basic_user"

                with patch("mlflow_oidc_auth.utils.request") as mock_request:
                    mock_request.authorization = AuthBasic()
                    self.assertEqual(get_username(), "basic_user")

                # missing username in basic auth
                class AuthBasicNone:
                    type = "basic"
                    username = None

                with patch("mlflow_oidc_auth.utils.request") as mock_request:
                    mock_request.authorization = AuthBasicNone()
                    with self.assertRaises(MlflowException):
                        get_username()
            # bearer token
            with patch("mlflow_oidc_auth.utils.session", {}):

                class AuthBearer:
                    type = "bearer"
                    token = "tok"

                mock_validate_token.return_value = {"email": "bearer_user"}
                with patch("mlflow_oidc_auth.utils.request") as mock_request:
                    mock_request.authorization = AuthBearer()
                    self.assertEqual(get_username(), "bearer_user")
            # no auth
            with patch("mlflow_oidc_auth.utils.session", {}):
                with patch("mlflow_oidc_auth.utils.request") as mock_request:
                    mock_request.authorization = None
                    with self.assertRaises(MlflowException):
                        get_username()

    @patch("mlflow_oidc_auth.utils._get_tracking_store")
    def test_get_experiment_id_experiment_name_not_found(self, mock_tracking_store):
        # experiment_name provided but not found
        with self.app.test_request_context("/?experiment_name=nonexistent_exp", method="GET"):
            mock_tracking_store().get_experiment_by_name.return_value = None
            with self.assertRaises(MlflowException) as cm:
                get_experiment_id()
            self.assertEqual(cm.exception.error_code, "INVALID_PARAMETER_VALUE")

    @patch("mlflow_oidc_auth.utils.store")
    @patch("mlflow_oidc_auth.utils.get_permission_from_store_or_default")
    def test_effective_experiment_permission(self, mock_get_permission_from_store_or_default, mock_store):
        with self.app.test_request_context():
            mock_get_permission_from_store_or_default.return_value = PermissionResult(
                Permission(name="perm", priority=1, can_read=True, can_update=True, can_delete=True, can_manage=True), "user"
            )
            result = effective_experiment_permission("exp_id", "user")
            self.assertTrue(result.permission.can_manage)
            self.assertEqual(result.type, "user")

    @patch("mlflow_oidc_auth.utils.store")
    @patch("mlflow_oidc_auth.utils.get_permission_from_store_or_default")
    def test_effective_registered_model_permission(self, mock_get_permission_from_store_or_default, mock_store):
        with self.app.test_request_context():
            mock_get_permission_from_store_or_default.return_value = PermissionResult(
                Permission(name="perm", priority=1, can_read=True, can_update=True, can_delete=True, can_manage=True), "user"
            )
            result = effective_registered_model_permission("model_name", "user")
            self.assertTrue(result.permission.can_manage)
            self.assertEqual(result.type, "user")

    @patch("mlflow_oidc_auth.utils.store")
    @patch("mlflow_oidc_auth.utils.get_permission_from_store_or_default")
    def test_effective_prompt_permission(self, mock_get_permission_from_store_or_default, mock_store):
        with self.app.test_request_context():
            mock_get_permission_from_store_or_default.return_value = PermissionResult(
                Permission(name="perm", priority=1, can_read=True, can_update=True, can_delete=True, can_manage=True), "user"
            )
            result = effective_prompt_permission("prompt_name", "user")
            self.assertTrue(result.permission.can_manage)
            self.assertEqual(result.type, "user")

    @patch("mlflow_oidc_auth.utils.store")
    @patch("mlflow_oidc_auth.utils.get_permission_from_store_or_default")
    def test_can_read_experiment(self, mock_get_permission_from_store_or_default, mock_store):
        with self.app.test_request_context():
            mock_get_permission_from_store_or_default.return_value = PermissionResult(
                Permission(name="perm", priority=1, can_read=True, can_update=False, can_delete=False, can_manage=False), "user"
            )
            self.assertTrue(can_read_experiment("exp_id", "user"))

    @patch("mlflow_oidc_auth.utils.store")
    @patch("mlflow_oidc_auth.utils.get_permission_from_store_or_default")
    def test_can_read_registered_model(self, mock_get_permission_from_store_or_default, mock_store):
        with self.app.test_request_context():
            mock_get_permission_from_store_or_default.return_value = PermissionResult(
                Permission(name="perm", priority=1, can_read=True, can_update=False, can_delete=False, can_manage=False), "user"
            )
            self.assertTrue(can_read_registered_model("model_name", "user"))

    def test_get_url_param(self):
        # URL param present
        with self.app.test_request_context("/test/123"):
            with patch("mlflow_oidc_auth.utils.request") as mock_request:
                mock_request.view_args = {"id": "123"}
                self.assertEqual(get_url_param("id"), "123")

        # URL param missing
        with self.app.test_request_context("/test"):
            with patch("mlflow_oidc_auth.utils.request") as mock_request:
                mock_request.view_args = {}
                with self.assertRaises(MlflowException) as cm:
                    get_url_param("id")
                self.assertEqual(cm.exception.error_code, "INVALID_PARAMETER_VALUE")

    def test_get_optional_url_param(self):
        # URL param present
        with self.app.test_request_context("/test/123"):
            with patch("mlflow_oidc_auth.utils.request") as mock_request:
                mock_request.view_args = {"id": "123"}
                self.assertEqual(get_optional_url_param("id"), "123")

        # URL param missing
        with self.app.test_request_context("/test"):
            with patch("mlflow_oidc_auth.utils.request") as mock_request:
                mock_request.view_args = {}
                self.assertIsNone(get_optional_url_param("id"))

    def test_get_model_name(self):
        # URL args
        with self.app.test_request_context("/model/test-model"):
            with patch("mlflow_oidc_auth.utils.request") as mock_request:
                mock_request.view_args = {"name": "test-model"}
                mock_request.args = {}
                mock_request.json = None
                self.assertEqual(get_model_name(), "test-model")

        # Query args
        with self.app.test_request_context("/?name=test-model", method="GET"):
            self.assertEqual(get_model_name(), "test-model")

        # JSON data
        with self.app.test_request_context("/", method="POST", json={"name": "test-model"}, content_type="application/json"):
            self.assertEqual(get_model_name(), "test-model")

        # Missing name
        with self.app.test_request_context("/", method="GET"):
            with patch("mlflow_oidc_auth.utils.request") as mock_request:
                mock_request.view_args = None
                mock_request.args = {}
                mock_request.json = None
                with self.assertRaises(MlflowException) as cm:
                    get_model_name()
                self.assertEqual(cm.exception.error_code, "INVALID_PARAMETER_VALUE")

    @patch("mlflow_oidc_auth.utils._get_tracking_store")
    def test_experiment_id_from_name(self, mock_tracking_store):
        # Experiment found
        mock_tracking_store().get_experiment_by_name.return_value.experiment_id = "123"
        result = _experiment_id_from_name("test-experiment")
        self.assertEqual(result, "123")

        # Experiment not found
        mock_tracking_store().get_experiment_by_name.return_value = None
        with self.assertRaises(MlflowException) as cm:
            _experiment_id_from_name("nonexistent-experiment")
        self.assertEqual(cm.exception.error_code, "INVALID_PARAMETER_VALUE")

    @patch("mlflow_oidc_auth.utils.store")
    @patch("mlflow_oidc_auth.utils.get_is_admin")
    @patch("mlflow_oidc_auth.utils.get_username")
    @patch("mlflow_oidc_auth.utils.make_forbidden_response")
    def test_check_admin_permission(self, mock_make_forbidden_response, mock_get_username, mock_get_is_admin, mock_store):
        with self.app.test_request_context():
            mock_get_username.return_value = "user"
            mock_get_is_admin.return_value = False
            mock_make_forbidden_response.return_value = "forbidden"

            @check_admin_permission
            def mock_func():
                return "success"

            self.assertEqual(mock_func(), "forbidden")

            # Admin allowed
            mock_get_is_admin.return_value = True
            self.assertEqual(mock_func(), "success")

    @patch("mlflow_oidc_auth.utils._get_model_registry_store")
    def test_fetch_all_registered_models(self, mock_model_store):
        # Single page
        mock_result = MagicMock()
        mock_result.__iter__ = MagicMock(return_value=iter([MagicMock(name="model1"), MagicMock(name="model2")]))
        mock_result.__len__ = MagicMock(return_value=2)
        mock_result.token = None
        mock_model_store().search_registered_models.return_value = mock_result

        result = fetch_all_registered_models()
        self.assertEqual(len(result), 2)

        # Multiple pages
        first_page = MagicMock()
        first_page.__iter__ = MagicMock(return_value=iter([MagicMock(name="model1")]))
        first_page.__len__ = MagicMock(return_value=1)
        first_page.token = "token123"

        second_page = MagicMock()
        second_page.__iter__ = MagicMock(return_value=iter([MagicMock(name="model2")]))
        second_page.__len__ = MagicMock(return_value=1)
        second_page.token = None

        mock_model_store().search_registered_models.side_effect = [first_page, second_page]

        result = fetch_all_registered_models()
        self.assertEqual(len(result), 2)

    @patch("mlflow_oidc_auth.utils.fetch_all_registered_models")
    def test_fetch_all_prompts(self, mock_fetch_models):
        mock_models = [MagicMock(name="prompt1"), MagicMock(name="prompt2")]
        mock_fetch_models.return_value = mock_models

        result = fetch_all_prompts()
        self.assertEqual(result, mock_models)
        mock_fetch_models.assert_called_once_with(filter_string="tags.`mlflow.prompt.is_prompt` = 'true'", max_results_per_page=1000)

    @patch("mlflow_oidc_auth.utils._get_model_registry_store")
    def test_fetch_registered_models_paginated(self, mock_model_store):
        mock_result = MagicMock()
        mock_model_store().search_registered_models.return_value = mock_result

        result = fetch_registered_models_paginated(filter_string="test_filter", max_results=100, order_by=["name"], page_token="token123")

        self.assertEqual(result, mock_result)
        mock_model_store().search_registered_models.assert_called_once_with(
            filter_string="test_filter", max_results=100, order_by=["name"], page_token="token123"
        )

    @patch("mlflow_oidc_auth.utils._get_tracking_store")
    def test_fetch_all_experiments(self, mock_tracking_store):
        # Single page
        mock_result = MagicMock()
        mock_result.__iter__ = MagicMock(return_value=iter([MagicMock(name="exp1"), MagicMock(name="exp2")]))
        mock_result.__len__ = MagicMock(return_value=2)
        mock_result.token = None
        mock_tracking_store().search_experiments.return_value = mock_result

        result = fetch_all_experiments()
        self.assertEqual(len(result), 2)

        # Multiple pages
        first_page = MagicMock()
        first_page.__iter__ = MagicMock(return_value=iter([MagicMock(name="exp1")]))
        first_page.__len__ = MagicMock(return_value=1)
        first_page.token = "token123"

        second_page = MagicMock()
        second_page.__iter__ = MagicMock(return_value=iter([MagicMock(name="exp2")]))
        second_page.__len__ = MagicMock(return_value=1)
        second_page.token = None

        mock_tracking_store().search_experiments.side_effect = [first_page, second_page]

        result = fetch_all_experiments()
        self.assertEqual(len(result), 2)

    @patch("mlflow_oidc_auth.utils._get_tracking_store")
    def test_fetch_experiments_paginated(self, mock_tracking_store):
        mock_result = MagicMock()
        mock_tracking_store().search_experiments.return_value = mock_result

        result = fetch_experiments_paginated(view_type=1, max_results=100, order_by=["name"], filter_string="test_filter", page_token="token123")

        self.assertEqual(result, mock_result)
        mock_tracking_store().search_experiments.assert_called_once_with(
            view_type=1, max_results=100, order_by=["name"], filter_string="test_filter", page_token="token123"
        )

    @patch("mlflow_oidc_auth.utils.fetch_all_experiments")
    @patch("mlflow_oidc_auth.utils.can_read_experiment")
    @patch("mlflow_oidc_auth.utils.get_username")
    def test_fetch_readable_experiments(self, mock_get_username, mock_can_read, mock_fetch_all):
        mock_get_username.return_value = "user"
        mock_exp1 = MagicMock()
        mock_exp1.experiment_id = "1"
        mock_exp2 = MagicMock()
        mock_exp2.experiment_id = "2"
        mock_fetch_all.return_value = [mock_exp1, mock_exp2]
        mock_can_read.side_effect = lambda exp_id, user: exp_id == "1"

        result = fetch_readable_experiments()
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0], mock_exp1)

    @patch("mlflow_oidc_auth.utils.fetch_all_registered_models")
    @patch("mlflow_oidc_auth.utils.can_read_registered_model")
    @patch("mlflow_oidc_auth.utils.get_username")
    def test_fetch_readable_registered_models(self, mock_get_username, mock_can_read, mock_fetch_all):
        mock_get_username.return_value = "user"
        mock_model1 = MagicMock()
        mock_model1.name = "model1"
        mock_model2 = MagicMock()
        mock_model2.name = "model2"
        mock_fetch_all.return_value = [mock_model1, mock_model2]
        mock_can_read.side_effect = lambda name, user: name == "model1"

        result = fetch_readable_registered_models()
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0], mock_model1)

    @patch("mlflow_oidc_auth.utils.store")
    def test_get_registered_model_permission_from_regex(self, mock_store):
        from mlflow_oidc_auth.entities import RegisteredModelRegexPermission

        regex_perms = [
            RegisteredModelRegexPermission(regex="test.*", permission="READ", priority=1, user_id=1),
            RegisteredModelRegexPermission(regex="prod.*", permission="MANAGE", priority=2, user_id=1),
        ]

        # Match found
        result = _get_registered_model_permission_from_regex(regex_perms, "test-model")
        self.assertEqual(result, "READ")

        # No match
        with self.assertRaises(MlflowException) as cm:
            _get_registered_model_permission_from_regex(regex_perms, "other-model")
        self.assertEqual(cm.exception.error_code, "RESOURCE_DOES_NOT_EXIST")

    @patch("mlflow_oidc_auth.utils.store")
    @patch("mlflow_oidc_auth.utils._get_tracking_store")
    def test_get_experiment_permission_from_regex(self, mock_tracking_store, mock_store):
        from mlflow_oidc_auth.entities import ExperimentRegexPermission

        mock_tracking_store().get_experiment.return_value.name = "test-experiment"

        regex_perms = [
            ExperimentRegexPermission(regex="test.*", permission="READ", priority=1, user_id=1),
            ExperimentRegexPermission(regex="prod.*", permission="MANAGE", priority=2, user_id=1),
        ]

        # Match found
        result = _get_experiment_permission_from_regex(regex_perms, "exp123")
        self.assertEqual(result, "READ")

        # No match
        mock_tracking_store().get_experiment.return_value.name = "other-experiment"
        with self.assertRaises(MlflowException) as cm:
            _get_experiment_permission_from_regex(regex_perms, "exp123")
        self.assertEqual(cm.exception.error_code, "RESOURCE_DOES_NOT_EXIST")

    @patch("mlflow_oidc_auth.utils.store")
    def test_get_registered_model_group_permission_from_regex(self, mock_store):
        from mlflow_oidc_auth.entities import RegisteredModelGroupRegexPermission

        regex_perms = [
            RegisteredModelGroupRegexPermission(id_=1, regex="test.*", permission="READ", priority=1, group_id=1),
            RegisteredModelGroupRegexPermission(id_=2, regex="prod.*", permission="MANAGE", priority=2, group_id=1),
        ]

        # Match found
        result = _get_registered_model_group_permission_from_regex(regex_perms, "test-model")
        self.assertEqual(result, "READ")

        # No match
        with self.assertRaises(MlflowException) as cm:
            _get_registered_model_group_permission_from_regex(regex_perms, "other-model")
        self.assertEqual(cm.exception.error_code, "RESOURCE_DOES_NOT_EXIST")

    @patch("mlflow_oidc_auth.utils.store")
    @patch("mlflow_oidc_auth.utils._get_tracking_store")
    def test_get_experiment_group_permission_from_regex(self, mock_tracking_store, mock_store):
        from mlflow_oidc_auth.entities import ExperimentGroupRegexPermission

        mock_tracking_store().get_experiment.return_value.name = "test-experiment"

        regex_perms = [
            ExperimentGroupRegexPermission(id_=1, regex="test.*", permission="READ", priority=1, group_id=1),
            ExperimentGroupRegexPermission(id_=2, regex="prod.*", permission="MANAGE", priority=2, group_id=1),
        ]

        # Match found
        result = _get_experiment_group_permission_from_regex(regex_perms, "exp123")
        self.assertEqual(result, "READ")

        # No match
        mock_tracking_store().get_experiment.return_value.name = "other-experiment"
        with self.assertRaises(MlflowException) as cm:
            _get_experiment_group_permission_from_regex(regex_perms, "exp123")
        self.assertEqual(cm.exception.error_code, "RESOURCE_DOES_NOT_EXIST")

    @patch("mlflow_oidc_auth.utils.store")
    def test_permission_sources_config(self, mock_store):
        # Test prompt sources config
        config = _permission_prompt_sources_config("model1", "user1")
        self.assertIn("user", config)
        self.assertIn("group", config)
        self.assertIn("regex", config)
        self.assertIn("group-regex", config)

        # Test experiment sources config
        config = _permission_experiment_sources_config("exp1", "user1")
        self.assertIn("user", config)
        self.assertIn("group", config)
        self.assertIn("regex", config)
        self.assertIn("group-regex", config)

        # Test registered model sources config
        config = _permission_registered_model_sources_config("model1", "user1")
        self.assertIn("user", config)
        self.assertIn("group", config)
        self.assertIn("regex", config)
        self.assertIn("group-regex", config)

    def test_get_request_param_run_id_fallback(self):
        # Test run_id fallback to run_uuid
        with self.app.test_request_context("/?run_uuid=test-uuid", method="GET"):
            result = get_request_param("run_id")
            self.assertEqual(result, "test-uuid")

    def test_get_request_param_post_json(self):
        # Test POST with JSON data
        with self.app.test_request_context("/", method="POST", json={"param": "value"}, content_type="application/json"):
            result = get_request_param("param")
            self.assertEqual(result, "value")

    def test_get_optional_request_param_post_json(self):
        # Test POST with JSON data
        with self.app.test_request_context("/", method="POST", json={"param": "value"}, content_type="application/json"):
            result = get_optional_request_param("param")
            self.assertEqual(result, "value")

    def test_get_experiment_id_view_args(self):
        # Test experiment_id from view_args
        with self.app.test_request_context("/test/123"):
            with patch("mlflow_oidc_auth.utils.request") as mock_request:
                mock_request.view_args = {"experiment_id": "123"}
                mock_request.args = {}
                mock_request.json = None
                result = get_experiment_id()
                self.assertEqual(result, "123")

    @patch("mlflow_oidc_auth.utils._get_tracking_store")
    def test_get_experiment_id_view_args_name(self, mock_tracking_store):
        # Test experiment_name from view_args
        mock_tracking_store().get_experiment_by_name.return_value.experiment_id = "456"
        with self.app.test_request_context("/test/exp-name"):
            with patch("mlflow_oidc_auth.utils.request") as mock_request:
                mock_request.view_args = {"experiment_name": "exp-name"}
                mock_request.args = {}
                mock_request.json = None
                result = get_experiment_id()
                self.assertEqual(result, "456")

    @patch("mlflow_oidc_auth.utils._get_tracking_store")
    def test_get_experiment_id_json_name(self, mock_tracking_store):
        # Test experiment_name from JSON
        mock_tracking_store().get_experiment_by_name.return_value.experiment_id = "789"
        with self.app.test_request_context("/", method="POST", json={"experiment_name": "exp-name"}, content_type="application/json"):
            result = get_experiment_id()
            self.assertEqual(result, "789")

    def test_get_username_bearer_missing_email(self):
        # Test bearer token without email
        with self.app.test_request_context():
            with patch("mlflow_oidc_auth.utils.session", {}):

                class AuthBearer:
                    type = "bearer"
                    token = "tok"

                with patch("mlflow_oidc_auth.utils.validate_token") as mock_validate:
                    mock_validate.return_value = {}  # No email field
                    with patch("mlflow_oidc_auth.utils.request") as mock_request:
                        mock_request.authorization = AuthBearer()
                        # The function returns None when email is missing, it doesn't raise an exception
                        result = get_username()
                        self.assertIsNone(result)

    def test_get_username_unknown_auth_type(self):
        # Test unknown auth type
        with self.app.test_request_context():
            with patch("mlflow_oidc_auth.utils.session", {}):

                class AuthUnknown:
                    type = "unknown"

                with patch("mlflow_oidc_auth.utils.request") as mock_request:
                    mock_request.authorization = AuthUnknown()
                    with self.assertRaises(MlflowException):
                        get_username()

    @patch("mlflow_oidc_auth.utils.store")
    @patch("mlflow_oidc_auth.utils.config")
    @patch("mlflow_oidc_auth.utils.get_permission")
    def test_get_permission_from_store_or_default_non_resource_exception(self, mock_get_permission, mock_config, mock_store):
        with self.app.test_request_context():
            mock_store_permission_user_func = MagicMock()
            mock_store_permission_user_func.side_effect = MlflowException("Other error", BAD_REQUEST)

            mock_config.PERMISSION_SOURCE_ORDER = ["user"]

            with self.assertRaises(MlflowException) as cm:
                get_permission_from_store_or_default({"user": mock_store_permission_user_func})
            self.assertEqual(cm.exception.error_code, "BAD_REQUEST")


if __name__ == "__main__":
    unittest.main()



from mlflow_oidc_auth.utils import get_request_param, get_username


def _username_is_sender():
    """Validate if the request username is the sender"""
    username = get_request_param("username")
    sender = get_username()
    return username == sender


def validate_can_get_user_token():
    return _username_is_sender()


def validate_can_read_user():
    return _username_is_sender()


def validate_can_create_user():
    # only admins can create user, but admins won't reach this validator
    return False


def validate_can_update_user_password():
    return _username_is_sender()


def validate_can_update_user_admin():
    # only admins can update, but admins won't reach this validator
    return False


def validate_can_delete_user():
    # only admins can delete, but admins won't reach this validator
    return False



from flask import jsonify
from mlflow.server.handlers import catch_mlflow_exception

from mlflow_oidc_auth.store import store
from mlflow_oidc_auth.utils import (
    can_manage_registered_model,
    check_prompt_permission,
    get_is_admin,
    get_request_param,
    get_username,
)


@catch_mlflow_exception
@check_prompt_permission
def create_group_prompt_permission(group_name: str, prompt_name: str):
    store.create_group_prompt_permission(group_name=group_name, name=prompt_name, permission=get_request_param("permission"))
    return jsonify({"message": "Group model permission has been created."})


@catch_mlflow_exception
@check_prompt_permission
def delete_group_prompt_permission(group_name: str, prompt_name: str):
    store.delete_group_prompt_permission(group_name, prompt_name)
    return jsonify({"message": "Group model permission has been deleted."})


@catch_mlflow_exception
@check_prompt_permission
def update_group_prompt_permission(group_name: str, prompt_name: str):
    store.update_group_prompt_permission(group_name, prompt_name, get_request_param("permission"))
    return jsonify({"message": "Group model permission has been updated."})


@catch_mlflow_exception
def get_group_prompts(group_name: str):
    models = store.get_group_prompts(group_name)
    if get_is_admin():
        return jsonify(
            [
                {
                    "name": model.name,
                    "permission": model.permission,
                }
                for model in models
            ]
        )
    current_user = store.get_user(get_username())
    return jsonify(
        [
            {
                "name": model.name,
                "permission": model.permission,
            }
            for model in models
            if can_manage_registered_model(
                model.name,
                current_user.username,
            )
        ]
    )



import re
from functools import wraps
from typing import Callable, Dict, List, NamedTuple, Optional
from flask import request, session
from mlflow.exceptions import MlflowException
from mlflow.protos.databricks_pb2 import BAD_REQUEST, INVALID_PARAMETER_VALUE, RESOURCE_DOES_NOT_EXIST, ErrorCode
from mlflow.server import app
from mlflow.server.handlers import _get_tracking_store, _get_model_registry_store
from mlflow.entities.model_registry import RegisteredModel
from mlflow.entities import Experiment
from mlflow.store.entities.paged_list import PagedList

from mlflow_oidc_auth.auth import validate_token
from mlflow_oidc_auth.config import config
from mlflow_oidc_auth.entities import (
    ExperimentGroupRegexPermission,
    ExperimentRegexPermission,
    RegisteredModelGroupRegexPermission,
    RegisteredModelRegexPermission,
)
from mlflow_oidc_auth.permissions import Permission, get_permission
from mlflow_oidc_auth.responses.client_error import make_forbidden_response
from mlflow_oidc_auth.store import store


def fetch_all_registered_models(
    filter_string: Optional[str] = None, order_by: Optional[List[str]] = None, max_results_per_page: int = 1000
) -> List[RegisteredModel]:
    """
    Fetch ALL registered models from the MLflow model registry using pagination.
    This ensures we get all models, not just the first page.

    Args:
        filter_string: Filter string for the search
        order_by: List of order by clauses
        max_results_per_page: Maximum number of results to fetch per page (default: 1000)

    Returns:
        List of ALL RegisteredModel objects
    """
    all_models = []
    page_token = None

    while True:
        result = _get_model_registry_store().search_registered_models(
            filter_string=filter_string, max_results=max_results_per_page, order_by=order_by, page_token=page_token
        )

        all_models.extend(result)

        # Check if there are more pages
        if hasattr(result, "token") and result.token:
            page_token = result.token
        else:
            break

    return all_models


def fetch_all_prompts(max_results_per_page: int = 1000) -> List[RegisteredModel]:
    """
    Fetch ALL registered models that are marked as prompts using pagination.
    This ensures we get all prompts, not just the first page.

    Args:
        max_results_per_page: Maximum number of results to fetch per page (default: 1000)

    Returns:
        List of ALL RegisteredModel objects that are prompts
    """
    filter_string = "tags.`mlflow.prompt.is_prompt` = 'true'"
    return fetch_all_registered_models(filter_string=filter_string, max_results_per_page=max_results_per_page)


def fetch_registered_models_paginated(
    filter_string: Optional[str] = None, max_results: int = 1000, order_by: Optional[List[str]] = None, page_token=None
) -> PagedList[RegisteredModel]:
    """
    Fetch registered models with pagination support.

    Args:
        filter_string: Filter string for the search
        max_results: Maximum number of results to return
        order_by: List of order by clauses
        page_token: Token for pagination

    Returns:
        PagedList of RegisteredModel objects
    """
    return _get_model_registry_store().search_registered_models(
        filter_string=filter_string,
        max_results=max_results,
        order_by=order_by,
        page_token=page_token,
    )


def _get_registered_model_permission_from_regex(regexes: List[RegisteredModelRegexPermission], model_name: str) -> str:
    for regex in regexes:
        if re.match(regex.regex, model_name):
            app.logger.debug(f"Regex permission found for model name {model_name}: {regex.permission} with regex {regex.regex} and priority {regex.priority}")
            return regex.permission
    raise MlflowException(
        f"model name {model_name}",
        error_code=RESOURCE_DOES_NOT_EXIST,
    )


def _get_experiment_permission_from_regex(regexes: List[ExperimentRegexPermission], experiment_id: str) -> str:
    experiment_name = _get_tracking_store().get_experiment(experiment_id).name
    for regex in regexes:
        if re.match(regex.regex, experiment_name):
            app.logger.debug(
                f"Regex permission found for experiment id {experiment_name}: {regex.permission} with regex {regex.regex} and priority {regex.priority}"
            )
            return regex.permission
    raise MlflowException(
        f"experiment id {experiment_id}",
        error_code=RESOURCE_DOES_NOT_EXIST,
    )


def _get_registered_model_group_permission_from_regex(regexes: List[RegisteredModelGroupRegexPermission], model_name: str) -> str:
    for regex in regexes:
        if re.match(regex.regex, model_name):
            app.logger.debug(
                f"Regex group permission found for model name {model_name}: {regex.permission} with regex {regex.regex} and priority {regex.priority}"
            )
            return regex.permission
    raise MlflowException(
        f"model name {model_name}",
        error_code=RESOURCE_DOES_NOT_EXIST,
    )


def _get_experiment_group_permission_from_regex(regexes: List[ExperimentGroupRegexPermission], experiment_id: str) -> str:
    experiment_name = _get_tracking_store().get_experiment(experiment_id).name
    for regex in regexes:
        if re.match(regex.regex, experiment_name):
            app.logger.debug(
                f"Regex group permission found for experiment id {experiment_name}: {regex.permission} with regex {regex.regex} and priority {regex.priority}"
            )
            return regex.permission
    raise MlflowException(
        f"experiment id {experiment_id}",
        error_code=RESOURCE_DOES_NOT_EXIST,
    )


def _permission_prompt_sources_config(model_name: str, username: str) -> Dict[str, Callable[[], str]]:
    return {
        "user": lambda model_name=model_name, user=username: store.get_registered_model_permission(model_name, user).permission,
        "group": lambda model_name=model_name, user=username: store.get_user_groups_registered_model_permission(model_name, user).permission,
        "regex": lambda model_name=model_name, user=username: _get_registered_model_permission_from_regex(
            store.list_prompt_regex_permissions(user), model_name
        ),
        "group-regex": lambda model_name=model_name, user=username: _get_registered_model_group_permission_from_regex(
            store.list_group_prompt_regex_permissions_for_groups_ids(store.get_groups_ids_for_user(user)), model_name
        ),
    }


def _permission_experiment_sources_config(experiment_id: str, username: str) -> Dict[str, Callable[[], str]]:
    return {
        "user": lambda experiment_id=experiment_id, user=username: store.get_experiment_permission(experiment_id, user).permission,
        "group": lambda experiment_id=experiment_id, user=username: store.get_user_groups_experiment_permission(experiment_id, user).permission,
        "regex": lambda experiment_id=experiment_id, user=username: _get_experiment_permission_from_regex(
            store.list_experiment_regex_permissions(user), experiment_id
        ),
        "group-regex": lambda experiment_id=experiment_id, user=username: _get_experiment_group_permission_from_regex(
            store.list_group_experiment_regex_permissions_for_groups_ids(store.get_groups_ids_for_user(user)), experiment_id
        ),
    }


def _permission_registered_model_sources_config(model_name: str, username: str) -> Dict[str, Callable[[], str]]:
    return {
        "user": lambda model_name=model_name, user=username: store.get_registered_model_permission(model_name, user).permission,
        "group": lambda model_name=model_name, user=username: store.get_user_groups_registered_model_permission(model_name, user).permission,
        "regex": lambda model_name=model_name, user=username: _get_registered_model_permission_from_regex(
            store.list_registered_model_regex_permissions(user), model_name
        ),
        "group-regex": lambda model_name=model_name, user=username: _get_registered_model_group_permission_from_regex(
            store.list_group_registered_model_regex_permissions_for_groups_ids(store.get_groups_ids_for_user(user)), model_name
        ),
    }


def get_url_param(param: str) -> str:
    """Extract a URL path parameter from Flask's request.view_args.

    Args:
        param: The name of the URL parameter to extract

    Returns:
        The parameter value

    Raises:
        MlflowException: If the parameter is not found in the URL path
    """
    view_args = request.view_args
    if not view_args or param not in view_args:
        raise MlflowException(
            f"Missing value for required URL parameter '{param}'. " "The parameter should be part of the URL path.",
            INVALID_PARAMETER_VALUE,
        )
    return view_args[param]


def get_optional_url_param(param: str) -> str | None:
    """Extract an optional URL path parameter from Flask's request.view_args.

    Args:
        param: The name of the URL parameter to extract

    Returns:
        The parameter value or None if not found
    """
    view_args = request.view_args
    if not view_args or param not in view_args:
        app.logger.debug(f"Optional URL parameter '{param}' not found in request path.")
        return None
    return view_args[param]


def get_request_param(param: str) -> str:
    if request.method == "GET":
        args = request.args
    elif request.method in ("POST", "PATCH", "DELETE"):
        args = request.json
    else:
        raise MlflowException(
            f"Unsupported HTTP method '{request.method}'",
            BAD_REQUEST,
        )

    if not args or param not in args:
        # Special handling for run_id
        if param == "run_id":
            return get_request_param("run_uuid")
        raise MlflowException(
            f"Missing value for required parameter '{param}'. " "See the API docs for more information about request parameters.",
            INVALID_PARAMETER_VALUE,
        )
    return args[param]


def get_optional_request_param(param: str) -> str | None:
    if request.method == "GET":
        args = request.args
    elif request.method in ("POST", "PATCH", "DELETE"):
        args = request.json
    else:
        raise MlflowException(
            f"Unsupported HTTP method '{request.method}'",
            BAD_REQUEST,
        )

    if not args or param not in args:
        app.logger.debug(f"Optional parameter '{param}' not found in request data.")
        return None
    return args[param]


def get_username() -> str:
    username = session.get("username")
    if username:
        app.logger.debug(f"Username from session: {username}")
        return username
    elif request.authorization is not None:
        if request.authorization.type == "basic":
            app.logger.debug(f"Username from basic auth: {request.authorization.username}")
            if request.authorization.username is not None:
                return request.authorization.username
            raise MlflowException("Username not found in basic auth.")
        if request.authorization.type == "bearer":
            username = validate_token(request.authorization.token).get("email")
            app.logger.debug(f"Username from bearer token: {username}")
            return username
    raise MlflowException("Authentication required. Please see documentation for details: ")


def get_is_admin() -> bool:
    return bool(store.get_user(get_username()).is_admin)


def _experiment_id_from_name(experiment_name: str) -> str:
    """
    Helper function to get the experiment ID from the experiment name.
    Raises an exception if the experiment does not exist.
    """
    experiment = _get_tracking_store().get_experiment_by_name(experiment_name)
    if experiment is None:
        raise MlflowException(
            f"Experiment with name '{experiment_name}' not found.",
            INVALID_PARAMETER_VALUE,
        )
    return experiment.experiment_id


def get_experiment_id() -> str:
    # Fastest: check view_args first
    if request.view_args:
        if "experiment_id" in request.view_args:
            return request.view_args["experiment_id"]
        elif "experiment_name" in request.view_args:
            return _experiment_id_from_name(request.view_args["experiment_name"])
    # Next: check args (GET)
    if request.args:
        if "experiment_id" in request.args:
            return request.args["experiment_id"]
        elif "experiment_name" in request.args:
            return _experiment_id_from_name(request.args["experiment_name"])
    # Last: check json (POST, PATCH, DELETE)
    if request.json:
        if "experiment_id" in request.json:
            return request.json["experiment_id"]
        elif "experiment_name" in request.json:
            return _experiment_id_from_name(request.json["experiment_name"])
    raise MlflowException(
        "Either 'experiment_id' or 'experiment_name' must be provided in the request data.",
        INVALID_PARAMETER_VALUE,
    )


def get_model_name() -> str:
    """
    Helper function to get the model name from the request.
    Raises an exception if the model name is not found.
    """
    if request.view_args and "name" in request.view_args:
        return request.view_args["name"]
    if request.args and "name" in request.args:
        return request.args["name"]
    if request.json and "name" in request.json:
        return request.json["name"]
    raise MlflowException(
        "Model name must be provided in the request data.",
        INVALID_PARAMETER_VALUE,
    )


class PermissionResult(NamedTuple):
    permission: Permission
    type: str


# TODO: check fi str can be replaced by Permission in function signature
def get_permission_from_store_or_default(PERMISSION_SOURCES_CONFIG: Dict[str, Callable[[], str]]) -> PermissionResult:
    """
    Attempts to get permission from store based on configured sources,
    and returns default permission if no record is found.
    Permissions are checked in the order defined in PERMISSION_SOURCE_ORDER.
    """
    for source_name in config.PERMISSION_SOURCE_ORDER:
        if source_name in PERMISSION_SOURCES_CONFIG:
            try:
                # Get the permission retrieval function from the configuration
                permission_func = PERMISSION_SOURCES_CONFIG[source_name]
                # Call the function to get the permission
                perm = permission_func()
                app.logger.debug(f"Permission found using source: {source_name}")
                return PermissionResult(get_permission(perm), source_name)
            except MlflowException as e:
                if e.error_code != ErrorCode.Name(RESOURCE_DOES_NOT_EXIST):
                    raise  # Re-raise exceptions other than RESOURCE_DOES_NOT_EXIST
                app.logger.debug(f"Permission not found using source {source_name}: {e}")
        else:
            app.logger.warning(f"Invalid permission source configured: {source_name}")

    # If no permission is found, use the default
    perm = config.DEFAULT_MLFLOW_PERMISSION
    app.logger.debug("Default permission used")
    return PermissionResult(get_permission(perm), "fallback")


def effective_experiment_permission(experiment_id: str, user: str) -> PermissionResult:
    """
    Attempts to get permission from store based on configured sources,
    and returns default permission if no record is found.
    Permissions are checked in the order defined in PERMISSION_SOURCE_ORDER.
    """
    return get_permission_from_store_or_default(_permission_experiment_sources_config(experiment_id, user))


def effective_registered_model_permission(model_name: str, user: str) -> PermissionResult:
    """
    Attempts to get permission from store based on configured sources,
    and returns default permission if no record is found.
    Permissions are checked in the order defined in PERMISSION_SOURCE_ORDER.
    """
    return get_permission_from_store_or_default(_permission_registered_model_sources_config(model_name, user))


def effective_prompt_permission(prompt_name: str, user: str) -> PermissionResult:
    """
    Attempts to get permission from store based on configured sources,
    and returns default permission if no record is found.
    Permissions are checked in the order defined in PERMISSION_SOURCE_ORDER.
    """
    return get_permission_from_store_or_default(_permission_prompt_sources_config(prompt_name, user))


def can_read_experiment(experiment_id: str, user: str) -> bool:
    permission = effective_experiment_permission(experiment_id, user).permission
    return permission.can_read


def can_read_registered_model(model_name: str, user: str) -> bool:
    permission = effective_registered_model_permission(model_name, user).permission
    return permission.can_read


def can_manage_experiment(experiment_id: str, user: str) -> bool:
    permission = effective_experiment_permission(experiment_id, user).permission
    return permission.can_manage


def can_manage_registered_model(model_name: str, user: str) -> bool:
    permission = effective_registered_model_permission(model_name, user).permission
    return permission.can_manage


def check_experiment_permission(f) -> Callable:
    @wraps(f)
    def decorated_function(*args, **kwargs):
        current_user = store.get_user(get_username())
        if not get_is_admin():
            app.logger.debug(f"Not Admin. Checking permission for {current_user.username}")
            experiment_id = get_experiment_id()
            if not can_manage_experiment(experiment_id, current_user.username):
                app.logger.warning(f"Change permission denied for {current_user.username} on experiment {experiment_id}")
                return make_forbidden_response()
        app.logger.debug(f"Change permission granted for {current_user.username}")
        return f(*args, **kwargs)

    return decorated_function


def check_registered_model_permission(f) -> Callable:
    @wraps(f)
    def decorated_function(*args, **kwargs):
        current_user = store.get_user(get_username())
        if not get_is_admin():
            app.logger.debug(f"Not Admin. Checking permission for {current_user.username}")
            model_name = get_model_name()
            if not can_manage_registered_model(model_name, current_user.username):
                app.logger.warning(f"Change permission denied for {current_user.username} on model {model_name}")
                return make_forbidden_response()
        app.logger.debug(f"Permission granted for {current_user.username}")
        return f(*args, **kwargs)

    return decorated_function


def check_prompt_permission(f) -> Callable:
    @wraps(f)
    def decorated_function(*args, **kwargs):
        current_user = store.get_user(get_username())
        if not get_is_admin():
            app.logger.debug(f"Not Admin. Checking permission for {current_user.username}")
            prompt_name = get_model_name()
            if not can_manage_registered_model(prompt_name, current_user.username):
                app.logger.warning(f"Change permission denied for {current_user.username} on prompt {prompt_name}")
                return make_forbidden_response()
        app.logger.debug(f"Permission granted for {current_user.username}")
        return f(*args, **kwargs)

    return decorated_function


def check_admin_permission(f) -> Callable:
    @wraps(f)
    def decorated_function(*args, **kwargs):
        current_user = store.get_user(get_username())
        if not get_is_admin():
            app.logger.warning(f"Admin permission denied for {current_user.username}")
            return make_forbidden_response()
        app.logger.debug(f"Admin permission granted for {current_user.username}")
        return f(*args, **kwargs)

    return decorated_function


def fetch_all_experiments(
    view_type: int = 1, max_results_per_page: int = 1000, order_by: Optional[List[str]] = None, filter_string: Optional[str] = None  # ACTIVE_ONLY
) -> List[Experiment]:
    """
    Fetch ALL experiments from the MLflow tracking store using pagination.
    This ensures we get all experiments, not just the first page.

    Args:
        view_type: ViewType for experiments (1=ACTIVE_ONLY, 2=DELETED_ONLY, 3=ALL)
        max_results_per_page: Maximum number of results to fetch per page (default: 1000)
        order_by: List of order by clauses
        filter_string: Filter string for the search

    Returns:
        List of ALL Experiment objects
    """
    all_experiments = []
    page_token = None

    while True:
        result = _get_tracking_store().search_experiments(
            view_type=view_type,
            max_results=max_results_per_page,
            order_by=order_by,
            filter_string=filter_string,
            page_token=page_token,
        )

        all_experiments.extend(result)

        # Check if there are more pages
        if hasattr(result, "token") and result.token:
            page_token = result.token
        else:
            break

    return all_experiments


def fetch_experiments_paginated(
    view_type: int = 1,
    max_results: int = 1000,
    order_by: Optional[List[str]] = None,
    filter_string: Optional[str] = None,
    page_token=None,  # ACTIVE_ONLY
) -> PagedList[Experiment]:
    """
    Fetch experiments with pagination support.

    Args:
        view_type: ViewType for experiments (1=ACTIVE_ONLY, 2=DELETED_ONLY, 3=ALL)
        max_results: Maximum number of results to return
        order_by: List of order by clauses
        filter_string: Filter string for the search
        page_token: Token for pagination

    Returns:
        PagedList of Experiment objects
    """
    return _get_tracking_store().search_experiments(
        view_type=view_type,
        max_results=max_results,
        order_by=order_by,
        filter_string=filter_string,
        page_token=page_token,
    )


def fetch_readable_experiments(
    view_type: int = 1,
    max_results_per_page: int = 1000,
    order_by: Optional[List[str]] = None,
    filter_string: Optional[str] = None,
    username: Optional[str] = None,  # ACTIVE_ONLY
) -> List[Experiment]:
    """
    Fetch ALL experiments that the user can read from the MLflow tracking store using pagination.
    This ensures we get all readable experiments, not just the first page.

    Args:
        view_type: ViewType for experiments (1=ACTIVE_ONLY, 2=DELETED_ONLY, 3=ALL)
        max_results_per_page: Maximum number of results to fetch per page (default: 1000)
        order_by: List of order by clauses
        filter_string: Filter string for the search
        username: Username to check permissions for (defaults to current user)

    Returns:
        List of Experiment objects that the user can read
    """
    if username is None:
        username = get_username()

    # Get all experiments matching the filter
    all_experiments = fetch_all_experiments(view_type=view_type, max_results_per_page=max_results_per_page, order_by=order_by, filter_string=filter_string)

    # Filter by permissions
    readable_experiments = [experiment for experiment in all_experiments if can_read_experiment(experiment.experiment_id, username)]

    return readable_experiments


def fetch_readable_registered_models(
    filter_string: Optional[str] = None, order_by: Optional[List[str]] = None, max_results_per_page: int = 1000, username: Optional[str] = None
) -> List[RegisteredModel]:
    """
    Fetch ALL registered models that the user can read from the MLflow model registry using pagination.
    This ensures we get all readable models, not just the first page.

    Args:
        filter_string: Filter string for the search
        order_by: List of order by clauses
        max_results_per_page: Maximum number of results to fetch per page (default: 1000)
        username: Username to check permissions for (defaults to current user)

    Returns:
        List of RegisteredModel objects that the user can read
    """
    if username is None:
        username = get_username()

    # Get all models matching the filter
    all_models = fetch_all_registered_models(filter_string=filter_string, order_by=order_by, max_results_per_page=max_results_per_page)

    # Filter by permissions
    readable_models = [model for model in all_models if can_read_registered_model(model.name, username)]

    return readable_models



from mlflow_oidc_auth.permissions import Permission
from mlflow_oidc_auth.utils import effective_registered_model_permission, get_username, get_model_name


def _get_permission_from_registered_model_name() -> Permission:
    model_name = get_model_name()
    username = get_username()
    return effective_registered_model_permission(model_name, username).permission


def validate_can_read_registered_model():
    return _get_permission_from_registered_model_name().can_read


def validate_can_update_registered_model():
    return _get_permission_from_registered_model_name().can_update


def validate_can_delete_registered_model():
    return _get_permission_from_registered_model_name().can_delete


def validate_can_manage_registered_model():
    return _get_permission_from_registered_model_name().can_manage



from datetime import datetime, timezone
from typing import Callable, List, Optional

from mlflow.exceptions import MlflowException
from mlflow.protos.databricks_pb2 import RESOURCE_ALREADY_EXISTS, RESOURCE_DOES_NOT_EXIST
from mlflow.utils.validation import _validate_username
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import Session
from werkzeug.security import check_password_hash, generate_password_hash

from mlflow_oidc_auth.db.models import SqlUser
from mlflow_oidc_auth.entities import User
from mlflow_oidc_auth.repository.utils import get_user


class UserRepository:
    def __init__(self, session_maker):
        self._Session: Callable[[], Session] = session_maker

    def create(self, username: str, password: str, display_name: str, is_admin: bool = False, is_service_account: bool = False) -> User:
        _validate_username(username)
        pwhash = generate_password_hash(password)
        with self._Session() as session:
            try:
                u = SqlUser(
                    username=username,
                    password_hash=pwhash,
                    display_name=display_name,
                    is_admin=is_admin,
                    is_service_account=is_service_account,
                )
                session.add(u)
                session.flush()
                return u.to_mlflow_entity()
            except IntegrityError as e:
                raise MlflowException(f"User '{username}' already exists: {e}", RESOURCE_ALREADY_EXISTS) from e

    def get(self, username: str) -> User:
        with self._Session() as session:
            u = session.query(SqlUser).filter(SqlUser.username == username).one_or_none()
            if u is None:
                raise MlflowException(f"User '{username}' not found", RESOURCE_DOES_NOT_EXIST)
            return u.to_mlflow_entity()

    def exist(self, username: str) -> bool:
        with self._Session() as session:
            return session.query(SqlUser).filter(SqlUser.username == username).first() is not None

    def list(self, is_service_account: bool = False, all: bool = False) -> List[User]:
        with self._Session() as session:
            q = session.query(SqlUser)
            if not all:
                q = q.filter(SqlUser.is_service_account == is_service_account)
            return [u.to_mlflow_entity() for u in q.all()]

    def update(
        self,
        username: str,
        password: Optional[str] = None,
        password_expiration: Optional[datetime] = None,
        is_admin: Optional[bool] = False,
        is_service_account: Optional[bool] = False,
    ) -> User:
        from werkzeug.security import generate_password_hash

        with self._Session() as session:
            user = get_user(session, username)
            if password is not None:
                user.password_hash = generate_password_hash(password)
            if password_expiration is not None:
                user.password_expiration = password_expiration
            if is_admin is not None:
                user.is_admin = is_admin
            if is_service_account is not None:
                user.is_service_account = is_service_account
            session.flush()
            return user.to_mlflow_entity()

    def delete(self, username: str) -> None:
        with self._Session() as session:
            user = get_user(session, username)
            if user is None:
                raise MlflowException(f"User '{username}' not found.")
            session.delete(user)
            session.flush()

    def authenticate(self, username: str, password: str) -> bool:
        with self._Session() as session:
            try:
                user = get_user(session, username)
                if user.password_expiration is not None:
                    if user.password_expiration.tzinfo is None:
                        user.password_expiration = user.password_expiration.replace(tzinfo=timezone.utc)
                    if user.password_expiration < datetime.now(timezone.utc):
                        return False
                return check_password_hash(getattr(user, "password_hash"), password)
            except MlflowException:
                return False



from unittest.mock import MagicMock, patch

from mlflow_oidc_auth.validators import registered_model


class DummyPermission:
    def __init__(self, can_read=False, can_update=False, can_delete=False, can_manage=False):
        self.can_read = can_read
        self.can_update = can_update
        self.can_delete = can_delete
        self.can_manage = can_manage


def _patch_permission(**kwargs):
    return patch(
        "mlflow_oidc_auth.validators.registered_model.effective_registered_model_permission",
        return_value=MagicMock(permission=DummyPermission(**kwargs)),
    )


def test__get_permission_from_registered_model_name():
    with patch("mlflow_oidc_auth.validators.registered_model.get_model_name", return_value="modelA"), patch(
        "mlflow_oidc_auth.validators.registered_model.get_username", return_value="alice"
    ), patch(
        "mlflow_oidc_auth.validators.registered_model.effective_registered_model_permission",
        return_value=MagicMock(permission=DummyPermission(can_read=True)),
    ):
        perm = registered_model._get_permission_from_registered_model_name()
        assert perm.can_read is True


def test_validate_can_read_registered_model():
    with patch("mlflow_oidc_auth.validators.registered_model.get_model_name", return_value="modelA"):
        with patch("mlflow_oidc_auth.validators.registered_model.get_username", return_value="alice"):
            with _patch_permission(can_read=True):
                assert registered_model.validate_can_read_registered_model() is True


def test_validate_can_update_registered_model():
    with patch("mlflow_oidc_auth.validators.registered_model.get_model_name", return_value="modelA"):
        with patch("mlflow_oidc_auth.validators.registered_model.get_username", return_value="alice"):
            with _patch_permission(can_update=True):
                assert registered_model.validate_can_update_registered_model() is True


def test_validate_can_delete_registered_model():
    with patch("mlflow_oidc_auth.validators.registered_model.get_model_name", return_value="modelA"):
        with patch("mlflow_oidc_auth.validators.registered_model.get_username", return_value="alice"):
            with _patch_permission(can_delete=True):
                assert registered_model.validate_can_delete_registered_model() is True


def test_validate_can_manage_registered_model():
    with patch("mlflow_oidc_auth.validators.registered_model.get_model_name", return_value="modelA"):
        with patch("mlflow_oidc_auth.validators.registered_model.get_username", return_value="alice"):
            with _patch_permission(can_manage=True):
                assert registered_model.validate_can_manage_registered_model() is True



from unittest import mock
from mlflow_oidc_auth import routes

"""
`routes` contains multiple routes definitions after refactoring.
This test ensures that all expected routes are present and properly defined.
"""


class TestRoutes:
    def test_routes_presented(self):
        assert all(
            route is not None
            for route in [
                # Basic auth routes
                routes.HOME,
                routes.LOGIN,
                routes.LOGOUT,
                routes.CALLBACK,
                routes.STATIC,
                routes.UI,
                routes.UI_ROOT,
                # User management routes
                routes.CREATE_ACCESS_TOKEN,
                routes.GET_CURRENT_USER,
                routes.CREATE_USER,
                routes.GET_USER,
                routes.UPDATE_USER_PASSWORD,
                routes.UPDATE_USER_ADMIN,
                routes.DELETE_USER,
                # List resources
                routes.LIST_EXPERIMENTS,
                routes.LIST_PROMPTS,
                routes.LIST_MODELS,
                routes.LIST_USERS,
                routes.LIST_GROUPS,
                # User permissions
                routes.USER_EXPERIMENT_PERMISSIONS,
                routes.USER_EXPERIMENT_PERMISSION_DETAIL,
                routes.USER_REGISTERED_MODEL_PERMISSIONS,
                routes.USER_REGISTERED_MODEL_PERMISSION_DETAIL,
                routes.USER_PROMPT_PERMISSIONS,
                routes.USER_PROMPT_PERMISSION_DETAIL,
                # Resource user permissions
                routes.EXPERIMENT_USER_PERMISSIONS,
                routes.EXPERIMENT_USER_PERMISSION_DETAIL,
                routes.REGISTERED_MODEL_USER_PERMISSIONS,
                routes.REGISTERED_MODEL_USER_PERMISSION_DETAIL,
                routes.PROMPT_USER_PERMISSIONS,
                routes.PROMPT_USER_PERMISSION_DETAIL,
                # User pattern permissions
                routes.USER_EXPERIMENT_PATTERN_PERMISSIONS,
                routes.USER_EXPERIMENT_PATTERN_PERMISSION_DETAIL,
                routes.USER_REGISTERED_MODEL_PATTERN_PERMISSIONS,
                routes.USER_REGISTERED_MODEL_PATTERN_PERMISSION_DETAIL,
                routes.USER_PROMPT_PATTERN_PERMISSIONS,
                routes.USER_PROMPT_PATTERN_PERMISSION_DETAIL,
                # Group permissions
                routes.GROUP_EXPERIMENT_PERMISSIONS,
                routes.GROUP_EXPERIMENT_PERMISSION_DETAIL,
                routes.GROUP_REGISTERED_MODEL_PERMISSIONS,
                routes.GROUP_REGISTERED_MODEL_PERMISSION_DETAIL,
                routes.GROUP_PROMPT_PERMISSIONS,
                routes.GROUP_PROMPT_PERMISSION_DETAIL,
                # Resource group permissions
                routes.EXPERIMENT_GROUP_PERMISSIONS,
                routes.EXPERIMENT_GROUP_PERMISSION_DETAIL,
                routes.REGISTERED_MODEL_GROUP_PERMISSIONS,
                routes.REGISTERED_MODEL_GROUP_PERMISSION_DETAIL,
                routes.PROMPT_GROUP_PERMISSIONS,
                routes.PROMPT_GROUP_PERMISSION_DETAIL,
                # Group pattern permissions
                routes.GROUP_EXPERIMENT_PATTERN_PERMISSIONS,
                routes.GROUP_EXPERIMENT_PATTERN_PERMISSION_DETAIL,
                routes.GROUP_REGISTERED_MODEL_PATTERN_PERMISSIONS,
                routes.GROUP_REGISTERED_MODEL_PATTERN_PERMISSION_DETAIL,
                routes.GROUP_PROMPT_PATTERN_PERMISSIONS,
                routes.GROUP_PROMPT_PATTERN_PERMISSION_DETAIL,
                # Group user permissions
                routes.GROUP_USER_PERMISSIONS,
            ]
        )



import pytest
from unittest.mock import patch, MagicMock
from flask import Flask, session, request, Response
from mlflow_oidc_auth.hooks.before_request import before_request_hook
from mlflow_oidc_auth import responses
from mlflow_oidc_auth.config import config

app = Flask(__name__)
app.secret_key = "test_secret_key"


@pytest.fixture
def client():
    with app.test_client() as client:
        yield client


def test_unprotected_route(client):
    with app.test_request_context(path="/health", method="GET"):
        assert before_request_hook() is None  # No response for unprotected routes


def test_basic_auth_failure(client):
    with app.test_request_context(path="/protected", method="GET", headers={"Authorization": "Basic invalid"}):
        with patch("mlflow_oidc_auth.hooks.before_request.authenticate_request_basic_auth", return_value=False), patch(
            "mlflow_oidc_auth.hooks.before_request.render_template", return_value=Response("Unauthorized", status=401)
        ):
            response = before_request_hook()
            assert response.status_code == 401  # type: ignore
            assert b"Unauthorized" in response.data  # type: ignore


def test_bearer_auth_failure(client):
    with app.test_request_context(path="/protected", method="GET", headers={"Authorization": "Bearer invalid"}):
        with patch("mlflow_oidc_auth.hooks.before_request.authenticate_request_bearer_token", return_value=False):
            response = before_request_hook()
            assert response.status_code == 401  # type: ignore


def test_session_redirect(client):
    with app.test_request_context(path="/protected", method="GET"):
        session.clear()
        with patch("mlflow_oidc_auth.hooks.before_request.config.AUTOMATIC_LOGIN_REDIRECT", True), patch(
            "mlflow_oidc_auth.hooks.before_request.url_for", return_value="/login"
        ):
            response = before_request_hook()
            assert response.status_code == 302  # type: ignore
            assert response.location.endswith("/login")  # type: ignore


def test_authorization_failure(client):
    with app.test_request_context(path="/protected", method="GET"):
        with patch("mlflow_oidc_auth.hooks.before_request.get_is_admin", return_value=False), patch(
            "mlflow_oidc_auth.hooks.before_request.BEFORE_REQUEST_VALIDATORS", {("/protected", "GET"): lambda: False}
        ), patch("mlflow_oidc_auth.hooks.before_request.render_template", return_value=Response("Forbidden", status=403)):
            response = before_request_hook()
            assert response.status_code == 403  # type: ignore
            assert b"Forbidden" in response.data  # type: ignore



from flask import jsonify, make_response
from mlflow.server.handlers import catch_mlflow_exception

from mlflow_oidc_auth.store import store
from mlflow_oidc_auth.utils import check_admin_permission, get_request_param


@catch_mlflow_exception
@check_admin_permission
def create_experiment_regex_permission(username: str):
    store.create_experiment_regex_permission(
        get_request_param("regex"),
        int(get_request_param("priority")),
        get_request_param("permission"),
        username,
    )
    return jsonify({"status": "success"}), 200


@catch_mlflow_exception
@check_admin_permission
def list_experiment_regex_permission(username: str):
    ep = store.list_experiment_regex_permissions(username=username)
    return make_response({"experiment_permission": [e.to_json() for e in ep]})


@catch_mlflow_exception
@check_admin_permission
def get_experiment_regex_permission(username: str, pattern_id: str):
    ep = store.get_experiment_regex_permission(
        username=username,
        id=int(pattern_id),
    )
    return make_response({"experiment_permission": ep.to_json()})


@catch_mlflow_exception
@check_admin_permission
def update_experiment_regex_permission(username: str, pattern_id: str):
    ep = store.update_experiment_regex_permission(
        regex=get_request_param("regex"),
        priority=int(get_request_param("priority")),
        permission=get_request_param("permission"),
        username=username,
        id=int(pattern_id),
    )
    return make_response({"experiment_permission": ep.to_json()})


@catch_mlflow_exception
@check_admin_permission
def delete_experiment_regex_permission(username: str, pattern_id: str):
    store.delete_experiment_regex_permission(
        username=username,
        id=int(pattern_id),
    )
    return make_response({"status": "success"})



import pytest
from unittest.mock import MagicMock, patch
from flask import Flask, Response
from mlflow.protos.service_pb2 import CreateExperiment, SearchExperiments
from mlflow.protos.model_registry_pb2 import CreateRegisteredModel, DeleteRegisteredModel, SearchRegisteredModels
from mlflow_oidc_auth.hooks.after_request import after_request_hook, AFTER_REQUEST_PATH_HANDLERS

app = Flask(__name__)


@pytest.fixture
def mock_response():
    response = MagicMock(spec=Response)
    response.status_code = 200
    response.json = {}
    response.data = b"{}"
    return response


@pytest.fixture
def mock_store():
    with patch("mlflow_oidc_auth.hooks.after_request.store") as mock_store:
        yield mock_store


@pytest.fixture
def mock_utils():
    with patch("mlflow_oidc_auth.hooks.after_request.get_username", return_value="test_user") as mock_username, patch(
        "mlflow_oidc_auth.hooks.after_request.get_is_admin", return_value=False
    ) as mock_is_admin:
        yield mock_username, mock_is_admin


def test_after_request_hook_no_handler(mock_response):
    with app.test_request_context(path="/unknown/path", method="GET", headers={"Content-Type": "application/json"}):
        result = after_request_hook(mock_response)
        assert result == mock_response


def test_delete_can_manage_registered_model_permission(mock_response, mock_store):
    with app.test_request_context(
        path="/api/2.0/mlflow/registered-models/delete",
        method="DELETE",
        json={"name": "test_model"},  # Send parameters in the body as JSON
        headers={"Content-Type": "application/json"},
    ):
        handler = AFTER_REQUEST_PATH_HANDLERS[DeleteRegisteredModel]
        with patch("mlflow_oidc_auth.utils.get_request_param", return_value="test_model"):
            handler(mock_response)
            mock_store.wipe_group_model_permissions.assert_called_once_with("test_model")
            mock_store.wipe_registered_model_permissions.assert_called_once_with("test_model")


def test_filter_search_experiments(mock_response, mock_store, mock_utils):
    handler = AFTER_REQUEST_PATH_HANDLERS[SearchExperiments]
    mock_response.json = {"experiments": [{"experiment_id": "123"}]}
    with app.test_request_context(path="/api/2.0/mlflow/experiments/search", method="POST", headers={"Content-Type": "application/json"}):
        with patch("mlflow_oidc_auth.hooks.after_request.can_read_experiment", side_effect=lambda exp_id, _: exp_id != "123"):
            handler(mock_response)
            assert len(mock_response.json["experiments"]) == 1


def test_filter_search_registered_models(mock_response, mock_store, mock_utils):
    handler = AFTER_REQUEST_PATH_HANDLERS[SearchRegisteredModels]
    mock_response.json = {"registered_models": [{"name": "test_model"}]}
    with app.test_request_context(path="/api/2.0/mlflow/registered-models/search", method="POST", headers={"Content-Type": "application/json"}):
        with patch("mlflow_oidc_auth.hooks.after_request.can_read_registered_model", side_effect=lambda name, _: name != "test_model"):
            handler(mock_response)
            assert len(mock_response.json["registered_models"]) == 1



import secrets

from flask import redirect, render_template, request, session, url_for
from mlflow.server import app

from mlflow_oidc_auth.auth import get_oauth_instance, process_oidc_callback
from mlflow_oidc_auth.config import config


def login():
    state = secrets.token_urlsafe(16)
    session["oauth_state"] = state
    oauth_instance = get_oauth_instance(app)
    if oauth_instance is None or oauth_instance.oidc is None:
        app.logger.error("OAuth instance or OIDC is not properly initialized")
        return "Internal Server Error", 500
    return oauth_instance.oidc.authorize_redirect(config.OIDC_REDIRECT_URI, state=state)


def logout():
    session.clear()
    if config.AUTOMATIC_LOGIN_REDIRECT:
        return render_template(
            "auth.html",
            username=None,
            provide_display_name=config.OIDC_PROVIDER_DISPLAY_NAME,
        )
    return redirect("/")


def callback():
    """Validate the state to protect against CSRF and handle login."""

    email, errors = process_oidc_callback(request, session)
    if errors:
        return render_template(
            "auth.html",
            username=None,
            provide_display_name=config.OIDC_PROVIDER_DISPLAY_NAME,
            error_messages=errors,
        )
    session["username"] = email
    return redirect(url_for("oidc_ui"))



from flask import jsonify
from mlflow.server.handlers import catch_mlflow_exception

from mlflow_oidc_auth.store import store
from mlflow_oidc_auth.utils import check_admin_permission


@catch_mlflow_exception
@check_admin_permission
def list_user_experiment_regex_permission(username: str):
    ep = store.list_experiment_regex_permissions(username=username)
    return jsonify([e.to_json() for e in ep]), 200


@catch_mlflow_exception
@check_admin_permission
def get_user_experiment_regex_permission(username: str, pattern_id: str):
    ep = store.get_experiment_regex_permission(id=int(pattern_id), username=username)
    if ep is None:
        return jsonify({"error": "Experiment regex permission not found"}), 404
    return jsonify(ep.to_json()), 200



import pytest
from flask import Flask, jsonify
from mlflow_oidc_auth.responses.client_error import (
    make_auth_required_response,
    make_forbidden_response,
    make_basic_auth_response,
)


@pytest.fixture(scope="module")
def test_app():
    app = Flask(__name__)
    with app.app_context():
        yield app


class TestClientErrorResponses:
    def test_make_auth_required_response(self, test_app):
        response = make_auth_required_response()
        assert response.status_code == 401
        assert response.get_json() == {"message": "Authentication required"}

    def test_make_forbidden_response(self, test_app):
        response = make_forbidden_response()
        assert response.status_code == 403
        assert response.get_json() == {"message": "Permission denied"}

    def test_make_forbidden_response_custom_message(self, test_app):
        custom_msg = {"message": "Custom permission denied message"}
        response = make_forbidden_response(custom_msg)
        assert response.status_code == 403
        assert response.get_json() == custom_msg

    def test_make_basic_auth_response(self, test_app):
        response = make_basic_auth_response()
        assert response.status_code == 401
        assert response.data.decode() == ("You are not authenticated. Please see documentation for details" "https://github.com/mlflow-oidc/mlflow-oidc-auth")
        assert response.headers["WWW-Authenticate"] == 'Basic realm="mlflow"'



import importlib
import os
import secrets

from dotenv import load_dotenv
from mlflow.server import app

load_dotenv()  # take environment variables from .env.
app.logger.setLevel(os.environ.get("LOG_LEVEL", "INFO"))


def get_bool_env_variable(variable, default_value):
    value = os.environ.get(variable, str(default_value))
    return value.lower() in ["true", "1", "t"]


class AppConfig:
    def __init__(self):
        self.DEFAULT_MLFLOW_PERMISSION = os.environ.get("DEFAULT_MLFLOW_PERMISSION", "MANAGE")
        self.SECRET_KEY = os.environ.get("SECRET_KEY", secrets.token_hex(16))
        self.OIDC_USERS_DB_URI = os.environ.get("OIDC_USERS_DB_URI", "sqlite:///auth.db")
        self.OIDC_GROUP_NAME = [group.strip() for group in os.environ.get("OIDC_GROUP_NAME", "mlflow").split(",")]
        self.OIDC_ADMIN_GROUP_NAME = os.environ.get("OIDC_ADMIN_GROUP_NAME", "mlflow-admin")
        self.OIDC_PROVIDER_DISPLAY_NAME = os.environ.get("OIDC_PROVIDER_DISPLAY_NAME", "Login with OIDC")
        self.OIDC_DISCOVERY_URL = os.environ.get("OIDC_DISCOVERY_URL", None)
        self.OIDC_GROUPS_ATTRIBUTE = os.environ.get("OIDC_GROUPS_ATTRIBUTE", "groups")
        self.OIDC_SCOPE = os.environ.get("OIDC_SCOPE", "openid,email,profile")
        self.OIDC_GROUP_DETECTION_PLUGIN = os.environ.get("OIDC_GROUP_DETECTION_PLUGIN", None)
        self.OIDC_REDIRECT_URI = os.environ.get("OIDC_REDIRECT_URI", None)
        self.OIDC_CLIENT_ID = os.environ.get("OIDC_CLIENT_ID", None)
        self.OIDC_CLIENT_SECRET = os.environ.get("OIDC_CLIENT_SECRET", None)
        self.AUTOMATIC_LOGIN_REDIRECT = get_bool_env_variable("AUTOMATIC_LOGIN_REDIRECT", False)
        self.OIDC_ALEMBIC_VERSION_TABLE = os.environ.get("OIDC_ALEMBIC_VERSION_TABLE", "alembic_version")
        self.PERMISSION_SOURCE_ORDER = [source.strip() for source in os.environ.get("PERMISSION_SOURCE_ORDER", "user,group,regex,group-regex").split(",")]

        # session
        self.SESSION_TYPE = os.environ.get("SESSION_TYPE", "cachelib")
        self.SESSION_PERMANENT = get_bool_env_variable("SESSION_PERMANENT", False)
        self.SESSION_KEY_PREFIX = os.environ.get("SESSION_KEY_PREFIX", "mlflow_oidc:")
        self.PERMANENT_SESSION_LIFETIME = os.environ.get("PERMANENT_SESSION_LIFETIME", 86400)
        if self.SESSION_TYPE:
            try:
                session_module = importlib.import_module(f"mlflow_oidc_auth.session.{(self.SESSION_TYPE).lower()}")
                app.logger.debug(f"Session module for {self.SESSION_TYPE} imported.")
                for attr in dir(session_module):
                    if attr.isupper():
                        setattr(self, attr, getattr(session_module, attr))
            except ImportError:
                app.logger.error(f"Session module for {self.SESSION_TYPE} could not be imported.")
        # cache
        self.CACHE_TYPE = os.environ.get("CACHE_TYPE", "FileSystemCache")
        if self.CACHE_TYPE:
            try:
                cache_module = importlib.import_module(f"mlflow_oidc_auth.cache.{(self.CACHE_TYPE).lower()}")
                app.logger.debug(f"Cache module for {self.CACHE_TYPE} imported.")
                for attr in dir(cache_module):
                    if attr.isupper():
                        setattr(self, attr, getattr(cache_module, attr))
            except ImportError:
                app.logger.error(f"Cache module for {self.CACHE_TYPE} could not be imported.")


config = AppConfig()



import os
import sys
from tempfile import mkstemp
from unittest.mock import patch, MagicMock
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from mlflow_oidc_auth.db.utils import migrate, migrate_if_needed


class TestMigrate:
    @patch("mlflow_oidc_auth.db.utils.upgrade")
    def test_migrate(self, mock_upgrade):
        engine = create_engine("sqlite:///:memory:")
        with sessionmaker(bind=engine)():
            migrate(engine, "head")

        mock_upgrade.assert_called_once()

    @patch("mlflow_oidc_auth.db.utils.MigrationContext")
    @patch("mlflow_oidc_auth.db.utils.ScriptDirectory")
    @patch("mlflow_oidc_auth.db.utils.upgrade")
    def test_migrate_if_needed_not_called_if_not_needed(self, mock_upgrade, mock_script_dir, mock_migration_context):
        script_dir_mock = MagicMock()
        script_dir_mock.get_current_head.return_value = "head"
        mock_script_dir.from_config.return_value = script_dir_mock
        mock_migration_context.configure.return_value.get_current_revision.return_value = "head"

        engine = create_engine("sqlite:///:memory:")
        with sessionmaker(bind=engine)():
            migrate_if_needed(engine, "head")

        mock_upgrade.assert_not_called()

    @patch("mlflow_oidc_auth.db.utils.MigrationContext")
    @patch("mlflow_oidc_auth.db.utils.ScriptDirectory")
    @patch("mlflow_oidc_auth.db.utils.upgrade")
    def test_migrate_if_needed_called_if_needed(self, mock_upgrade, mock_script_dir, mock_migration_context):
        script_dir_mock = MagicMock()
        script_dir_mock.get_current_head.return_value = "head"
        mock_script_dir.from_config.return_value = script_dir_mock
        mock_migration_context.configure.return_value.get_current_revision.return_value = "not_head"

        engine = create_engine("sqlite:///:memory:")
        with sessionmaker(bind=engine)():
            migrate_if_needed(engine, "head")

        mock_upgrade.assert_called_once()


class TestModifiedVersionTable:
    @patch.dict(os.environ, {"OIDC_ALEMBIC_VERSION_TABLE": "alembic_modified_version"})
    def test_different_alembic_version_table(self):
        # Force reload of the config module
        if "mlflow_oidc_auth.config" in sys.modules:
            del sys.modules["mlflow_oidc_auth.config"]

        # Create temporary file
        _, db_file = mkstemp()

        engine = create_engine(f"sqlite:///{db_file}")
        with sessionmaker(bind=engine)() as f:
            migrate(engine, "head")

        tables = []
        with engine.begin() as conn:
            connection = conn.connection

            connection = f.connection().connection
            cursor = connection.cursor()

            query = "SELECT name FROM sqlite_schema WHERE type ='table' AND name NOT LIKE 'sqlite_%'"
            tables = [x[0] for x in cursor.execute(query).fetchall()]

        # Delete the temp file again
        os.unlink(db_file)

        # Do the asserts
        assert "alembic_modified_version" in tables
        assert "alembic_version" not in tables


class TestDefaultVersionTable:
    def test_default_alembic_table(self):
        # Force reload of the config module
        if "mlflow_oidc_auth.config" in sys.modules:
            del sys.modules["mlflow_oidc_auth.config"]

        # Create temporary file
        _, db_file = mkstemp()

        # If we have residual config options in environment, clean it
        if "OIDC_ALEMBIC_VERSION_TABLE" in os.environ:
            del os.environ["OIDC_ALEMBIC_VERSION_TABLE"]

        engine = create_engine(f"sqlite:///{db_file}")
        with sessionmaker(bind=engine)() as f:
            migrate(engine, "head")

        tables = []
        with engine.begin() as conn:
            connection = conn.connection

            connection = f.connection().connection
            cursor = connection.cursor()

            query = "SELECT name FROM sqlite_schema WHERE type ='table' AND name NOT LIKE 'sqlite_%'"
            tables = [x[0] for x in cursor.execute(query).fetchall()]

        # Remove the temp file again
        os.unlink(db_file)

        # Do the assert
        assert "alembic_version" in tables
